import zipfile
import requests
import random
import time
import numpy as np
from io import BytesIO
import zipfile
import os
import matplotlib.pyplot as plt
import pandas as pd
import plotly.express as px
import plotly.colors as pc
from plotly.graph_objs.layout import Colorscale
import json
import copy

import threed_optix.utils.api as au
import threed_optix.utils.general as gu
import threed_optix.utils.vars as v

class RayDataFrame(pd.DataFrame):
    def __init__(self):
        raise TypeError("Cannot directly create an instance of RayDataFrame.")

    @classmethod
    def _new(cls, ray_table_url, maps_url, setup_object):
        ray_table = au._map_ray_table(ray_table_url, maps_url)
        ray_table.setup = setup_object
        return ray_table

class RayTable:
    """
    Contains information about simulation results.

    Attributes:
        _url (str): The URL of the results file.
        _data (pd.DataFrame): The data of the simulation results.
    """

    ##Magic methods and constructors
    def __init__(self):
        raise TypeError("Cannot directly create an instance of Result.")

    @classmethod
    def _new(cls, ray_table_url, maps_url, setup_object):
        """
        Creates a new results object based on the URL.

        Args:
            url (str): The URL of the results file.

        Returns:
            Result: The created Result object.
        """
        ray_table = object.__new__(cls)
        ray_table._ray_table_url = ray_table_url
        ray_table._data = None
        ray_table._maps_url = maps_url
        ray_table.setup = setup_object
        return ray_table

    def __str__(self):
        string = json.dumps(
            {
                "setup": self.setup.name,
                "setup id": self.setup.id,
                "data url": self._ray_table_url,
            },
            indent = 4,
        )
        return string

    ##Properties
    @property
    def data(self):
        """
        Property to access the data of the simulation results.
        If accessed, calls get_data.

        Returns:
            pd.DataFrame: The data of the simulation results.
        """
        #We will have to do something here as well- process the data by maps url by some sort of an unkown way that eliav will say
        return self._get_data()

    ##'Private' Utils

    def _get_data(self):
        """
        If data is not None, returns data.
        If data is None, fetches data and returns data.

        Returns:
            pd.DataFrame: The data of the simulation results.
        """
        if self._data is None:
            self._data = au._map_ray_table(self._ray_table_url, self._maps_url)
        return self._data

    ##Main 'Public' Methods
    def to_csv(self, path: str, **kwargs):
        """
        Saves the data to a CSV file.

        Args:
            path (str): The path to save the CSV file.
        """
        self.data.to_csv(path, **kwargs)

class Analysis:

    _added = False
    _urls = []
    fail_message = None
    raw_results = {}
    results = {}

    def __init__(self,
                 surface,
                 resolution,
                 rays,
                 name,
                 fast = False,
                 ):
        self.surface = surface
        self.name = name
        self.rays = {laser.id: num_rays for laser, num_rays in rays.items()}
        self.resolution = {'x': resolution[0], 'y': resolution[1]}
        self.fast = fast
        self.id = Analysis.generate_id()
        num_rays = self.rays.values()

        Analysis.assert_analysis_params(fast = fast, name = name, num_rays = num_rays)

    @classmethod
    def assert_analysis_params(cls, fast, name, num_rays):
        fast_valid_names = v.FAST_ANALYSIS_NAMES
        advanced_valid_names = v.ANALYSIS_NAMES
        if fast:
            assert name in fast_valid_names, f"Valid names for fast analysis are {fast_valid_names}"
            assert all([num <= 200 for num in num_rays]), f'Number of rays must be less than 200 for fast analysis'
        else:
            assert name in advanced_valid_names, f"Valid names for advanced analysis are {advanced_valid_names}"


    @classmethod
    def _new(cls, surface, resolution, num_rays, name, type, id):
        analysis = object.__new__(cls)
        analysis.surface = surface
        analysis.resolution = resolution
        analysis.rays = num_rays
        analysis.name = name
        analysis.fast = False if type == '1' else True
        analysis.id = id
        analysis._added = True
        return analysis

    @property
    def wls(self):
        return self._analysis_wls()

    @classmethod
    def generate_id(cls):
        # Generate a base36 string representation of the current timestamp.
        int_time = int(time.time())

        enc_36_time = np.base_repr(int_time, 36)
        randint = np.base_repr(random.randint(0, 36**5), 36)[2:5]
        id_ = enc_36_time + randint

        return id_

    def _analysis_wls(self):
        analysis_wls = []
        setup = self.surface._part._setup
        lasers_ids = self.rays.keys()
        laser_objects = [setup[laser] for laser in lasers_ids]
        for laser in laser_objects:
            wls_dicts = laser.data['light_source']['wavelengths_data']
            wls = [wls_dict['wavelength'] for wls_dict in wls_dicts]
            analysis_wls += wls
        analysis_wls = sorted(list(set(analysis_wls)))
        return analysis_wls

    def _extract_file(self,url, destination):
        response = requests.get(url)
        # Check if the request was successful
        if response.status_code == 200:
            # Create a BytesIO object from the response content
            zip_data = BytesIO(response.content)

            # Create a ZipFile object from the BytesIO object
            with zipfile.ZipFile(zip_data, 'r') as zip_ref:
                # Extract the contents of the zip file to a folder (you can change the destination folder)
                zip_ref.extractall(destination)

        file_name = url.split('/')[-1].replace(f'{self.id}_', '').replace('.zip', '')
        file_path = f'{destination}/{file_name}'
        return file_path

    def _unpack(self):
        setup = self.surface._part._setup.id
        destination_path = f'.analysis-files/{setup}/{self.surface.id}/{self.id}'
        if not os.path.exists(destination_path):
            os.makedirs(destination_path)
        for url in self._urls:
            self._extract_file(url, destination_path)

        return destination_path

    def _read_file(self, file_path):
        with open(file_path, 'rb') as f:
            content = f.read()
        header_format = v.ANALYSIS_HEADER_DTYPES
        header = np.frombuffer(content[:v.HEADER_BYTES], dtype=header_format)
        data = np.frombuffer(content[v.HEADER_BYTES:], dtype=v.ANALYSIS_MATRIX_DTYPES)
        return header, data

    def process_results(self):
        directory = self._unpack()
        for file in os.listdir(directory):
            file_results = {}
            file_path = f'{directory}/{file}'
            headers_nums, data = self._read_file(file_path)
            headers = gu.process_headers(headers_nums)
            file_results['metadata'] = headers
            data_matrices = data.reshape(len(self.wls), self.resolution['x'], self.resolution['y'])
            file_results['data'] = {}
            for i, wl in enumerate(self.wls):
                matrix = data_matrices[i]
                file_results['data'][wl] = matrix
            file_name = file.split('.')[0]
            self.raw_results[file_name] = file_results
        polarized_dict = gu.reorganize_analysis_results_dict(self.raw_results.values())
        self.results = polarized_dict
        return self.results

    def __str__(self):
        string = json.dumps(
            {
                "name": self.name,
                "id": self.id,
                "resolution": self.resolution,
                "rays": self.rays,
                "surface": self.surface.id,
                "part": self.surface._part.id,
            },
            indent=4,
        )

        return string

    def __eq__(self, other):

        if not isinstance(other, Analysis):
            return False

        is_rays_equal = self.rays == other.rays
        is_name_equal = self.name == other.name
        is_resolution_x_equal = self.resolution['x'] == other.resolution['x']
        is_resolution_y_equal = self.resolution['y'] == other.resolution['y']
        is_surface_equal = self.surface.id == other.surface.id

        if is_rays_equal and is_name_equal and is_surface_equal and is_resolution_x_equal and is_resolution_y_equal:
            return True

        return False


    def show(self, figsize = (20, 20)):

        if not self.results:
            raise Exception('Analysis was not run yet')

        num_polarizations = len(self.results)
        num_wavelengths = list(self.results.values())[0]['metadata']['num_wavelengths']
        fig = plt.figure(constrained_layout=True, figsize=figsize)

        subfigs = fig.subfigures(nrows=num_polarizations, ncols=1)

        if num_polarizations == 1:
            subfigs = [subfigs]

        for polarization, subfig in zip(self.results.keys(), subfigs):

            subfig.suptitle(f'Polarization {polarization}')
            axs = subfig.subplots(nrows=1, ncols=num_wavelengths)

            if num_wavelengths == 1:
                axs = [axs]

            for wavelength, ax in zip(self.results[polarization]['data'].keys(), axs):
                ax.imshow(self.results[polarization]['data'][wavelength])
                ax.set_title(f'Wavelength {wavelength} nm')

        plt.show()


        return None


    def show_interactive(self, polarizations=None, wavelengths=None):
        if not polarizations:
            polarizations = list(self.results.keys())
        if not wavelengths:
            wavelengths = list(self.results[polarizations[0]]['data'].keys())

        colorscale = Colorscale({'sequential': pc.sequential.Plasma})
        for polarization in polarizations:
            for wavelength in wavelengths:
                data = self.results[polarization]['data'][wavelength]
                fig = px.imshow(data, title=f'Polarization {polarization} Wavelength {wavelength} nm')
                fig.update_layout(height=600, width=600, colorscale=colorscale)
                fig.show()

    def copy(self):
        copied = copy.deepcopy(self)
        copied.id = Analysis.generate_id()
        copied._added = False
        return copied



import os
import pickle
import pandas as pd
import time
import random
import numpy as np
import requests
from io import BytesIO
import zipfile

import threed_optix.utils.math as mu
import threed_optix.utils.general as gu
import threed_optix.utils.api as au
import threed_optix.parts as pt
import threed_optix.utils.vars as v
import threed_optix.analyses as an

class Setup(object):
    """
    Used to manage the simulation setup and its parts.

    Attributes:
        name (str): The name of the setup.
        id (str): The id of the setup.
        _opt (str): The opt file of the setup.
        _api (ThreedOptixAPI): The pointer to the API that created it.
        parts (list): The list of part objects (or classes that inherit from Part).
    """

    ##Magic methods and constructors
    def __init__(self):
        raise TypeError("Cannot directly create an instance of Setup.")

    @classmethod
    def _new(cls, _api, setup_tuple: tuple):
        """
        Creates a setup with API pointer, id, and name only.

        Args:
            _api (ThreedOptixAPI): The API pointer.
            info_json (dict): Information JSON.

        Returns:
            Setup: The newly created Setup object.
        """
        setup = object.__new__(cls)
        setup.id = setup_tuple[0]
        setup.name = setup_tuple[1]
        setup._api = _api
        setup._opt = None
        setup._parts = None
        return setup

    def __len__(self) -> int:
        """
        Returns the number of parts that the setup has.

        Returns:
            int: The number of parts in the setup.
        """
        return len(self.parts)

    def __iter__(self):
        """
        Allows iterating through the parts of the setup.

        Yields:
            Part: The next Part object.
        """
        return iter(self.parts)

    def __contains__(self, part):
        """
        Allows checking if a part is in the setup.

        Args:
            part: The part to check.

        Returns:
            bool: True if the part is in the setup, False otherwise.
        """

        if not isinstance(part, (pt.Part, str)):
            raise TypeError(f"Invalid part type {type(part)}. Must be Part or part id.")

        if isinstance(part, pt.Part):
            for p in self:
                if p.id == part.id:
                    return True

        if isinstance(part, str):
            for p in self:
                if p.id == part:
                    return True

        return False

    def __getitem__(self, key):
        """
        Allows getting parts by index.

        Args:
            key: The index of the part.

        Returns:
            Part: The requested Part object.
        """
        if isinstance(key, int):
            return self.parts[key]
        if isinstance(key, str):
            for part in self:
                if part.id == key:
                    return part
            raise KeyError(f"Part with id {key} not found in the setup.")
        raise TypeError(f"Invalid key type {type(key)}. Must be part index or id.")

    def __str__(self) -> str:
        string = f"Setup {self.name} ({self.id}) with {len(self)} parts:\n"
        for part in self:
            string += f"  - {part.label} ({part.id})\n"
        return string

    def __call__(self, changes):
        raise NotImplementedError("This method is not implemented yet.")

    ##Properties
    @property
    def parts(self):
        """
        Property to access the list of part objects.

        Returns:
            list: The list of Part objects.
        """
        if self._parts is None:
            self._get_parts()
        parts = self._parts
        return parts

    ##'Private' Utils
    def _get_parts(self):
        parts = self._api._get_setup_parts(self.id).get(self.id).get('parts')
        self._parts = [pt.Part._new(_setup=self, id=part['id']) for part in parts]
        return None

    def _get_part(self, part_id):
        part = self._api._get_part(part_id, self.id)
        return part

    ##'Public' Utils
    def get(self,
            part_label: str,
            all: bool = False):
        """
        Returns the part object with the specified label.

        Args:
            part_label (str): The label of the part.

        Returns:
            Part: The requested Part object.
        """
        if all:
            parts = []
            for part in self:
                if part.label == part_label:
                    parts.append(part)
            return parts

        for part in self:
            if part.label == part_label:
                return part

        return None

    def at(self, location: tuple):
        """
        Returns the closest part object to the specified location in the global coordinate system.

        Args:
            location (tuple): The global coordinates (x, y, z) of the location.

        Returns:
            Part: The closest Part object.
        """

        distances = []

        for part in self:
            distance = mu._3d_distance(location, part._pose._position)
            distances.append(distance)

        min_value = min(distances)
        min_index = distances.index(min_value)
        min_part = self[min_index]
        print(f'Closest part is {min_part.label} at the location {min_part._pose}')
        return min_part

    def save(self, file_path: str):
        """
        Saves the object to a re-creatable pickle file.

        Args:
            file_path (str): The path to the pickle file.
        """
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)
        print(f"Setup saved to {file_path}.")

    @classmethod
    def load(cls, file_path: str, api):
        """
        Loads the object from the pickle file path.

        Args:
            file_path (str): The path to the pickle file.
            _api (ThreedOptixAPI): The API instance.

        Returns:
            Setup: The loaded Setup object.
        """
        with open(file_path, 'rb') as f:
            setup = pickle.load(f)
        if api is not None:
            setup._api = api
        print(f"Setup loaded from {file_path}.")
        return setup

    ##Main 'Public' Methods
    def plot(self):
        """
        Plots the setup to visualize its configuration.
        """
        raise NotImplementedError("This method is not implemented yet.")

    def restore(self, to_restore):
        if isinstance(to_restore, pt.Part):
            part = to_restore
            r = self.update_part(part)
            if not r[0]:
                raise Exception(r[1])

            current_part = self[part.id]
            current_part._data = part._data
            current_part._pose = part._pose
            return None

        elif isinstance(to_restore, Setup):
            setup = to_restore
            for part in setup:
                self.restore(part)
                return None
        raise ValueError(f"Invalid type {type(to_restore)} to restore: must be Part or Setup.")

    def update_part(self, part):
        return self._api.update_part(self.id, part)

class Job:
    """
    Contains information about a batch job.

    Attributes:
        _api (ThreedOptixAPI): The API pointer.
        _setup (Setup): The setup pointer.
        __base_url (str): The base URL for results.
        _prefix (str): The prefix of the results.
        __num_changes (int): Number of requested changes.
    """

    ##Magic methods and constructors
    def __init__(self):
        raise TypeError("Cannot directly create an instance of Job.")

    @classmethod
    def _from_json(cls, json_: dict, _api, _setup):
        """
        Creates the Job instance from the response JSON.

        Args:
            json (dict): The JSON response.

        Returns:
            Job: The created Job instance.
        """
        job = object.__new__(cls)
        job._url = json_['url'].replace('$', '')
        job._setup = _setup
        job._api = _api
        job._prefix = json_['simulation_file_prefix']
        job._num_changes = json_['number_of_changes']
        job.results = [an.RayTable._new(job._url.format(index = index), _setup) for index in range(job._num_changes)]
        return job

    def __str__(self):
        string = f'Job with {self._num_changes} changes and prefix {self._prefix} at {self._url}.\nBased on setup {self._setup.name} ({self._setup.id})'
        return string

    def __getitem__(self, index):
        """
        Gets the URL of the analysis at the specified location.

        Args:
            index: The index of the analysis.

        Returns:
            str: The URL of the analysis.
        """
        return self.results[index]

    def __len__(self) -> int:
        """
        Returns the number of changes.

        Returns:
            int: The number of changes.
        """
        return self._num_changes

    def __iter__(self):
        """
        Allows iterations over the URLs.

        Yields:
            str: The next URL.
        """
        return iter(self.results)

    ##Main 'Public' Methods
    def pull(self, index, filepath, **kwargs):
        """
        Gets the results of the analysis at the specified index.

        Args:
            index: The index of the analysis.

        Returns:
            dict: The results of the analysis.
        """
        result = self[index]
        result.to_csv(filepath, **kwargs)

    def status(self):
        """
        Returns a status report of the job.

        Returns:
            str: The status report.
        """
        raise NotImplementedError("This method is not implemented yet.")



import threading
import pickle
import math
from typing import Any
import requests
import pandas as pd
import numpy as np
import random
import time
import requests
import zipfile
from io import BytesIO
import os

import threed_optix.utils.api as au
import threed_optix.utils.vars as v
import threed_optix.utils.math as mu
import threed_optix.utils.general as gu
import threed_optix.analyses as an
import threed_optix.simulations as sim

class ThreedOptixAPI:
    """
    Used to manage the communication with the server of 3DOptix.

    Args:
        api_key (str): The API key used for authentication.

    Attributes:
        api_key (str): The API key of the user.
        jobs (list): A list of Job objects for batch jobs.
        setups (list): The list of setups of the user.
    """
    setups = None
    ##Magic methods
    def __init__(self, 
                 api_key: str,
                 verbose = True
                 ):
        self.api_key = api_key
        self.jobs = []

        assert self.is_up(), v.SERVER_DOWN_MESSAGE
        assert self.is_key_valid(), v.INVALID_KEY_MESSAGE

        if verbose:
            welcome_thread = threading.Thread(target=au._welcome)
            welcome_thread.start()

        setups_thread = threading.Thread(target=self._initialize_setups)
        setups_thread.start()

        # Wait for both threads to finish
        if verbose:
            welcome_thread.join()

        setups_thread.join()
        if self.setups is None:
            raise Exception("Could not initialize setups, check your API key")

    def __contains__(self,
                     item
                     ):
        """
        Allows checking if a setup name or id is in the API.

        Args:
            item: The setup name, id, or object.

        Returns:
            bool: True if the setup exists, False otherwise.
        """
        contains = False
        if isinstance(item, str):
            for setup in self:
                if setup.id == item:
                    contains = True
                    break
        elif isinstance(item, sim.Setup):
            for setup in self:
                if setup.id == item.id:
                    contains = True
                    break
        return contains

    def __len__(self) -> int:
        """
        Returns the number of setups.

        Returns:
            int: The number of setups.
        """
        return len(self.setups)

    def __iter__(self):
        """
        Allows iterating through the setups.

        Yields:
            Setup: The next Setup object.
        """
        return iter(self.setups)

    def __str__(self) -> str:
        string = f"API with {len(self)} setups:\n"
        for setup in self:
            string += f"  - {setup.name} ({setup.id})\n"
        return string

    def __getitem__(self, key):
        """
        Allows accessing API setups by index, id, or name.

        Args:
            key: The index or id of the setup.

        Returns:
            Setup: The corresponding Setup object.
        """
        if isinstance(key, int):
            return self.setups[key]
        elif isinstance(key, str):
            for setup in self:
                if setup.id == key:
                    return setup
            raise KeyError(f"Setup with id {key} not found.")
        raise TypeError(f"Invalid key type {type(key)}. Must be setup index or id.")

    ##'Private' Utils
    def _initialize_setups(self):
        self.setups = self._get_setups_info()
        if self.setups is not None:
            self.setups = [sim.Setup._new(_api=self, setup_tuple=setup) for setup in self.setups]
        return self.setups

    def _get_setups_info(self) -> list:
        """
        Returns a list of Setup objects that are associated with the user.

        Returns:
            list: A list of Setup objects.
        """
        data, message = au._get_setups(self.api_key)
        if data is not None:
            infos = []
            for info_json in data['setups']:
                infos.append((info_json['id'], info_json['name']))
            return infos
        #else:
            #raise Exception(message)
        return None

    def _get_setup_parts(self, setup_id: str) -> list:
        parts = au._get_setup(setup_id, self.api_key)[0]
        return parts

    def _get_part(self, part_id: str, setup_id) -> dict:
        part = au._get_part(setup_id,part_id,  self.api_key)[0]
        return part

    def _extract_setup_object(self, setup):

        if isinstance(setup, str):

            if setup in self:
                setup_object = self[setup]
                return setup_object

            elif self.get(setup) is not None:
                setup_object = self.get(setup)
                return setup_object

            else:
                raise Exception(f"Setup with id or name {setup} not found.")

        if isinstance(setup, sim.Setup):
            return setup

        if isinstance(setup, int):
            return self.setups[setup]

        raise TypeError(f"Invalid setup type {type(setup)}. Must be Setup object, name, index or id.")

    ##'Public' Utils
    def get(self,
            setup_name: str,
            all: bool = False
            ):
        """
        Returns the Setup object with the specified name.

        Args:
            setup_name (str): The name of the setup.
            all (bool): If True, returns all setups with the specified name. else, returns the first one

        Returns:
            Setup: The Setup object.
        """
        if all:
            setups = []
            for setup in self:
                if setup.name == setup_name:
                    setups.append(setup)
            return setups

        else:
            setup = None
            for s in self.setups:
                if s.name == setup_name:
                    setup = s
                    return setup

    def is_up(self) -> bool:
        """
        Calls _healthcheck and returns a boolean indicating if the server is up.

        Returns:
            bool: True if the server is up, False otherwise.
        """
        return au._healthcheck()[0]

    def is_key_valid(self) -> bool:
        return True

    def update_part(self, setup_id, part):
        success, message = au._set_part(setup_id, part.id, part.data, self.api_key)
        if not success:
            raise Exception(message)
        return (success, message)

    ##Main 'Public' Methods
    def get_setups(self):
        """
        Returns a list of Setup objects that are associated with the user.

        Returns:
            list: A list of Setup objects.
        """
        return self.setups

    def run(self,
            setup
            ):
        """
        Sends the opt file to the 3DOptix server and returns a Results object.

        Args:
            setup (Setup): The setup to run.

        Returns:
            Results: The Results object.
        """
        setup_object = self._extract_setup_object(setup)

        data, message = au._run_simulation(setup_object.id, self.api_key)

        if data == None:
            raise Exception(message)

        if data['results'][0]['error']['code'] != 0:
            raise Exception(v.SIMULATION_ERROR.format(message = data['results'][0]['error']['message']))


        if data is not None:
            ray_table = data['results'][0]['data']['ray_table']
            maps_url = data['maps_url']
            return an.RayTable._new(ray_table, maps_url, setup_object)
        else:
            raise Exception(message)

    def run_async(self, setup):
        setup_object = self._extract_setup_object(setup)
        data, message = au._run_simulation(setup_object.id, self.api_key, is_sync = False)
        if data == None:
            raise Exception(message)
        if data['results'][0]['error']['code'] != 0:
            raise Exception(v.SIMULATION_ERROR.format(message = data['results'][0]['error']['message']))

        if data is not None:
            ray_table = data['results'][0]['data']['ray_table']
            maps_url = data['maps_url']
            return an.RayTable._new(ray_table, maps_url, setup_object)
        else:
            raise Exception(message)

    def run_batch(self,
                  setup,
                  configuration: dict
                  ):
        """
        Sends the setup id with the configuration to 3DOptix server and returns a Job object.
        Job object will be appended to self.jobs list.

        Args:
            setup (Setup): The setup to run.
            configuration (dict): The configuration for the batch job.

        Returns:
            Job: The Job object.
        """
        response = au._run_batch(setup.id, configuration, self.api_key)
        if response[0] is not None:
            json_ = response[0]
            json_['number_of_changes'] = configuration['number_of_changes']
            json_['simulation_file_prefix'] = configuration['simulation_file_prefix']
            job = sim.Job._from_json(response[0], _api = self, _setup = setup)
            job._url = job._url.replace('$', '')
            self.jobs.append(job)
            return job
        else:
            raise Exception(response[1])

    def run_analysis(self, analysis, auto_add = False, force = False):
        successful, failed =  self.run_analyses([analysis], auto_add = auto_add, force = force)
        if len(successful) == 1:
            return True
        return False

    def run_analyses(self, analyses, auto_add = False, force = False):

        ids = [analysis.id for analysis in analyses]
        setup_id = analyses[0].surface._part._setup.id

        response = {'success': [], 'failed': []}

        if not all([analysis.surface._part._setup.id == setup_id for analysis in analyses]):
            raise Exception(v.ANALYSES_NOT_SAME_SETUP_ERROR)

        if not auto_add and force:
            raise Exception("Force argument can only be used with auto_add argument")

        if auto_add:
            added_successfully = []
            added_failed = []

            for analysis in analyses:
                added = self.add_analysis(analysis, force = force)
                if added:
                    added_successfully.append((analysis, analysis.id))
                else:
                    print(f"Analysis {analysis.id} was failed to be added")
                    added_failed.append((analysis, analysis.id))
            response['added'] = added_successfully
            response['failed'] += added_failed
            ids = [analysis[1] for analysis in added_successfully]


        ran_successfully = []
        ran_failed = []

        if not all([analysis._added for analysis in analyses]):
            not_added = [analysis.id if analysis._added else None for analysis in analyses]
            raise Exception(v.ANALYSES_ADD_ERROR.format(not_added = not_added))

        data, message =  au._run_analyses(setup_id, self.api_key, ids)

        if data == None:
            raise Exception(message)

        for i, analysis in enumerate(analyses):
            is_successful = data['results'][i]['error']['code'] == 0
            ran_successfully.append((analysis, analysis.id))
            if is_successful:
                analysis_datos = data['results'][i]['data']['analysis']
                for data in analysis_datos:
                    url = data['url']
                    analysis._urls.append(url)
                analysis.results = analysis.process_results()
            else:
                ran_failed.append((analysis, analysis.id))
        gu.print_completed_failed(ran_successfully, ran_failed, message = 'Ran')
        return ran_successfully, ran_failed

    def add_analysis(self, analysis, force = False):

        if analysis.id in [a.id for a in analysis.surface.analyses]:
            analysis._added = True
            return analysis._added

        if not force:
            is_duplicated = False
            duplicated = []

            for existing_analysis in analysis.surface.analyses:
                if analysis == existing_analysis:
                    is_duplicated = True
                    duplicated.append(analysis.id)
            if is_duplicated:
                raise Exception(v.ANALYSES_DUPLICATED_ERROR.format(duplicated = ", ".join(duplicated)))

        data = {
            "surface_id": analysis.surface.id,
            "analyses_data": [
                {
                    "id": analysis.id ,
                    "name": analysis.name,
                    "num_rays": analysis.rays,
                    "resolution": analysis.resolution,
                    "type": 0 if analysis.fast == True else 1,
                }
                ]
            }
        part = analysis.surface._part
        response = au._add_analyses(part._setup.id, part.id, data, self.api_key)
        if response[1] == 'analyses successfully added':
            analysis._added = True
            analysis.surface.analyses.append(analysis)
            return analysis._added
        raise Exception(response[1])



import threed_optix.analyses as an
import threed_optix.utils.math as mu
import copy
import json

class Part:

    _data = None
    _surfaces = None
    ##Magic methods and constructors
    def __init__(self):
        raise TypeError("Cannot directly create an instance of Part.")

    @classmethod
    def _new(cls, _setup, id):
        """
        Creates a new part from the opt_part at the specified index within the setup.

        Args:
            _setup (Setup): The setup object.
            _index (int): The index of the part within the setup.

        Returns:
            Part: The newly created Part object.
        """

        part = object.__new__(cls)
        part._setup = _setup
        part.id = id
        part._pose = _Pose._new(part)
        return part

    def __str__(self) -> str:
        string = json.dumps(
            {
                "label": self.label,
                "id": self.id,
                "position": self.pose.vector[:3],
                "rotation": self.pose.vector[3:],
            },
            indent = 4
            )
        return string

    def __iter__(self):
        """
        Allows iterating through the surfaces of the part.

        Yields:
            Surface: The next Surface object.
        """
        return iter(self.surfaces)

    def __getitem__(self, key):
        """
        Allows getting surfaces by index.

        Args:
            key: The index of the surface.

        Returns:
            Surface: The requested Surface object.
        """
        if isinstance(key, int):
            return self.surfaces[key]
        if isinstance(key, str):
            for surface in self:
                if surface.id == key:
                    return surface
            raise KeyError(f"Surface with id {key} not found in the part.")
        raise TypeError(f"Invalid key type {type(key)}. Must be surface index or id.")

    ##Properties
    @property
    def data(self):
        if self._data is None:
            self._data = self._setup._get_part(self.id).get(self.id)
        data = self._data
        return data

    @property
    def pose(self):
        return self._pose

    @property
    def label(self):
        return self.data['label']

    @property
    def material(self):
        return self.data['material']

    @property
    def surfaces(self):
        if self._surfaces is None:
            self._surfaces = [Surface._new(self, surface) for surface in self.data.get('surfaces', []) if 'name' in surface]
        return self._surfaces

    ##'Public' Utils
    def get(self, name):
        """
        Returns the surface object with the specified name.

        Args:
            name (str): The name of the surface.

        Returns:
            Surface: The requested Surface object.
        """
        for surface in self:
            if surface.name == name:
                return surface
        return None

    ##Main 'Public' Methods
    def change_pose(self, vector, radians = False):

        assert len(vector) == 6, "Pose vector must be of length 6."

        if radians:
            vector[3:] = [mu.rad_to_deg(x) for x in vector[3:]]

        if self._data is None:
            self._data = self._setup._get_part(self.id).get(self.id)

        self.backup_pose = copy.deepcopy(self._data['pose'])

        self._data['pose']['rotation'] = vector[3:]
        self._data['pose']['position'] = vector[:3]

        self.sync()

    def sync(self):
        r = self._setup.update_part(self)
        if not r[0]:
            self._data['pose'] = self.backup_pose
            raise Exception(f"Error updating part: {r[1]}")

        for part in self._setup:
            if part._data is not None and part != self:
                part._data = None

    def copy(self):
        return copy.deepcopy(self)

class Detector(Part):

    _size = None

    def __init__(self):
        raise TypeError("Cannot directly create an instance of Detector.")

    @property
    def size(self):
        if self._size is None:
            self._size = self._data['size']
        return self._size

    def change_size(self, size):
        if self._data is None:
            self._data = self._setup._get_part(self.id).get(self.id)
        self._data['size'] = size
        self.sync()

class LightSource(Part):

    _light_source = None

    def __init__(self):
        raise TypeError("Cannot directly create an instance of LightSource.")

    @property
    def _light_source(self):
        if self._light_source is None:
            self._light_source = self.data['light_source']
        return self._light_source

class Surface:
    def __init__(self):
        raise TypeError("Cannot directly create an instance of Surface.")

    @classmethod
    def _new(cls, _part, _data):
        surface = object.__new__(cls)
        surface.name = _data['name']
        surface.id = _data['id']
        surface._part = _part
        surface.analyses = [an.Analysis._new(surface = surface, **analysis) for analysis in _data['analyses']]
        return surface

    def __str__(self):
        string = json.dumps(
            {
            "name": self.name,
            "id": self.id,
            "part id": self._part.id,
            "part label": self._part.label
            },
            indent = 4
            )
        return string

class _Pose:

    vector = None
    ##Magic methods and constructors

    def __init__(self):
        raise TypeError("Cannot directly create an instance of _Pose.")

    @classmethod
    def _new(cls, _part):
        pose = object.__new__(cls)
        pose._part = _part
        return pose

    @property
    def vector(self):
        return self._part._data['pose']['position'] + self._part._data['pose']['rotation']

    def __str__(self):
        string = json.dumps(
            {
                "position": self.vector[:3],
                "rotation": self.vector[3:],
                "part id": self._part.id,
                "part label": self._part.label
            },
            indent = 4
        )
        return string



import requests
from colorama import init, Fore, Style
import time
import json
import math
import numpy as np
from threed_optix.utils import vars as v
import pandas as pd
import threed_optix.utils.math as mu
import copy


def verify_response(response):
    if int(response.status_code) not in v.VALID_RESPONSE_CODES:
        raise Exception(f'Error: {response.status_code} - {response.text}')
    return None


#General Purpose
def _healthcheck():
    """
    Checks if the server is up and well.
    """
    url = f'{v.API_URL}/healthcheck'

    r = requests.get(url)
    verify_response(r)
    r = r.json()
    return (r['status'] == 'SUCCESS', r['message'])

#API Endpoints
##General
def _get(
    endpoint: str,
    api_key: str,
    #payload: dict = {}
    ):
    """
    Sends a GET request to the endpoint with the API key.

    Args:
        endpoint (str): The endpoint to send the request to.
        api_key (str): The API key.

    Returns:
        dict: The response JSON.
    """
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key}
    r = requests.get(url, headers=headers)
    verify_response(r)
    r = r.json()
    return (r['data'] if 'status' in r and r['status'] == 'SUCCESS' else None, r['message'])

def _put(endpoint: str, api_key: str, json_data: dict = None):
    """
    Sends a PUT request to the endpoint.

    Args:
        endpoint (str): The endpoint to send the request to.
        json_data (dict): The JSON payload for the request.
        api_key (str): The API key.
    """
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-Key': api_key, 'Content-Type': 'application/json'}
    r = requests.put(url, headers=headers, json=json_data)
    verify_response(r)
    r = r.json()
    return (r['data'] if 'status' in r and r['status'] == 'SUCCESS' else None, r['message'])

def _set(endpoint, data, api_key):
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key}

    r = requests.post(url, headers=headers, json=data)
    verify_response(r)
    r = r.json()
    return ('status' in r and r['status'] == 'SUCCESS', r['message'])

def _post(endpoint, data, api_key):
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key}

    r = requests.post(url, headers=headers, json=data)
    verify_response(r)
    r = r.json()
    return ('status' in r and r['status'] == 'SUCCESS', r['message'])

##Setups
def _get_setups(api_key: str):
    """
    Returns the list of setup names and ids of the user.

    Args:
        api_key (str): The API key.

    Returns:
        list: A list of setup names and ids.
    """
    endpoint = v.GET_SETUPS_ENDPOINT
    response = _get(endpoint, api_key)

    return response

def _get_setup(
    setup_id: str,
    api_key: str
    ):
    """
    Returns the opt file for the specified setup id.

    Args:
        setup_id (str): The setup id.
        api_key (str): The API key.

    Returns:
        str: The opt file content.
    """

    endpoint = v.GET_SETUP_ENDPOINT.format(setup_id=setup_id)

    return _get(endpoint, api_key)
    #return _dummy_opt()

##Parts
def _get_part(setup_id: str, part_id: str, api_key: str):
    endpoint = v.GET_PART_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    part =  _get(endpoint, api_key)
    return part

def _set_part(setup_id, part_id, data, api_key):
    data_copy = copy.deepcopy(data)
    data_copy['pose']['rotation'] =[mu.deg_to_rad(x) for x in data_copy['pose']['rotation']]
    endpoint = v.SET_PART_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    r = _set(endpoint, data_copy, api_key)
    return r

##Run Simulations
def _run_async(setup_id, api_key):
    endpoint =  v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)
    json_data = {
        "gpu_type": v.GPU_TYPE,
        "is_sync": True,
        "snellius_version": v.SNELLIUS_VERSION
    }
    r = _put(endpoint = endpoint, json_data=json_data, api_key= api_key)
    return r

def _run_batch(
    setup_id: str,
    configuration: dict,
    api_key: str):
    """
    Puts the batch run request.

    Args:
        setup_id (str): The setup id.
        configuration (dict): The batch configuration.
        api_key (str): The API key.
    """
    endpoint = v.PUT_BATCH_CHANGES_ENDPOINT.format(setup_id=setup_id)
    return _put(endpoint = endpoint, api_key= api_key, json_data = configuration)

def _run_simulation(setup_id, api_key, is_sync = True):
    endpoint =  v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)
    json_data = {
        "gpu_type": v.GPU_TYPE,
        "is_sync": is_sync,
        "snellius_version": v.SNELLIUS_VERSION
    }
    r = _put(endpoint = endpoint, json_data=json_data, api_key= api_key)
    return r

##Analyses
def _add_analyses(setup_id, part_id, data, api_key):
    endpoint = v.POST_ADD_ANALYSIS_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    return _put(endpoint = endpoint, json_data=data, api_key = api_key)

def _run_analyses(setup_id: str, api_key: str, analyses_ids: str):

    endpoint = v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)

    json_data = {
        "gpu_type": v.GPU_TYPE,
        "is_sync": True,
        "analyses_to_run": analyses_ids,
        "snellius_version": v.SNELLIUS_VERSION
    }

    return _put(endpoint = endpoint, json_data = json_data, api_key = api_key)


#Others
def _set_api_url(url, are_you_sure=False):

    if not are_you_sure:
        raise Exception(v.SET_API_URL_WARNING)

    print(v.SET_API_URL_WARNING)
    print('Previous API URL was', v.API_URL)
    print('Setting API URL to', url)
    v.API_URL = url

def _welcome():

    init(autoreset=True)
    color = Fore.WHITE
    print('******************')
    for i, char in enumerate(v.WELCOME_MESSAGE):
        print(f"{color}{char}", end="")
        time.sleep(0.02)  # Adjust the delay for the desired speed

    print(Style.RESET_ALL)  # Reset the style after the rainbow effect
    print('******************')

def _map_ray_row(row, map):
    row['surface_id'] = map['surfaces'][int(row['hit_surface_idx'])]
    row['wavelength'] = map['wavelengths'][int(row['wavelength_idx'])]
    row['light_source'] = map['sources'][int(row['source_idx'])]
    del row['hit_surface_idx']
    del row['wavelength_idx']
    del row['source_idx']
    return row

def _map_ray_table(rays_url, maps_url):
    rays_df = pd.read_csv(rays_url)
    rays_df = rays_df.copy()
    maps_json = requests.get(maps_url).json()
    rays_df = rays_df.apply(lambda row: _map_ray_row(row, maps_json), axis=1)
    return rays_df
